using System.Net;
using System.Net.Sockets;
using System.Text;
using DotNext;
using DotNext.Threading;

public struct UDPData
{
    public DateTime datetime;
    public string addr;
    public int port;
    public byte[] data;
}

/// <summary>
/// UDP Server
/// </summary>
public class UDPServer
{
    private int listenPort;
    private UdpClient listener;
    private IPEndPoint groupEP;
    private Dictionary<string, List<UDPData>> udpData = new Dictionary<string, List<UDPData>>();
    private AsyncReaderWriterLock udpDataLock = new AsyncReaderWriterLock(1);

    /// <summary>
    /// Construct a udp server with fixed port
    /// </summary>
    /// <param name="port"> Device UDP Port </param>
    /// <returns> UDPServer class </returns>
    public UDPServer(int port)
    {
        // Construction
        listenPort = port;
        try
        {
            listener = new UdpClient(listenPort);
            groupEP = new IPEndPoint(IPAddress.Any, listenPort);
        }
        catch (Exception e)
        {
            Console.WriteLine(e.ToString());
            throw new ArgumentException(
                $"Not currect port num: {port}",
                nameof(port)
                );
        }
    }

    /// <summary>
    /// Find UDP Receive Data According to ip address
    /// </summary>
    /// <param name="ipAddr"> IP Address</param>
    /// <param name="rwTimeout"> Read and Write Wait for Milliseconds </param>
    /// <returns>UDP Data</returns>
    public Optional<UDPData> FindData(string ipAddr, int rwTimeout = 1000)
    {
        UDPData? data = null;
        using (udpDataLock.AcquireWriteLock(TimeSpan.FromMilliseconds(rwTimeout)))
        {
            if (udpData.ContainsKey(ipAddr) && udpData[ipAddr].Count > 0)
            {
                data = udpData[ipAddr][0];
                udpData[ipAddr].RemoveAt(0);
            }
        }

        if (data == null)
        {
            return Optional.None<UDPData>();
        }
        else
        {
            return Optional.Some((UDPData)data);
        }
    }

    /// <summary>
    /// Find UDP Receive Data Async
    /// </summary>
    /// <param name="ipAddr">[TODO:parameter]</param>
    /// <param name="timeout">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Optional<UDPData>> FindDataAsync(string ipAddr, int timeout = 1000)
    {
        UDPData? data = null;
        using (await udpDataLock.AcquireWriteLockAsync(TimeSpan.FromMilliseconds(1000)))
        {
            if (udpData.ContainsKey(ipAddr) && udpData[ipAddr].Count > 0)
            {
                data = udpData[ipAddr][0];
                udpData[ipAddr].RemoveAt(0);
            }
        }

        if (data == null)
        {
            return Optional.None<UDPData>();
        }
        else
        {
            return Optional.Some((UDPData)data);
        }
    }

    private void ReceiveHandler(IAsyncResult res)
    {
        var remoteEP = new IPEndPoint(IPAddress.Any, listenPort);
        byte[] bytes = listener.EndReceive(res, ref remoteEP);
        var nowtime = DateTime.Now;

        // Handle RemoteEP
        string remoteStr = "Unknown";
        if (remoteEP is not null)
        {
            var remoteAddress = remoteEP.Address.ToString();
            var remotePort = remoteEP.Port;
            // Record UDP Receive Data
            if (udpData.ContainsKey(remoteAddress))
            {
                var listData = udpData[remoteAddress];
                listData.Add(new UDPData()
                {
                    addr = remoteAddress,
                    port = remotePort,
                    data = bytes,
                    datetime = nowtime,
                });
            }
            else
            {
                udpData.Add(remoteAddress, new List<UDPData>([new UDPData(){
                    addr = remoteAddress,
                    port = remotePort,
                    data = bytes,
                    datetime = nowtime,
                }]));
            }

            remoteStr = $"{remoteAddress}:{remotePort}";
        }

        // Handle Package
        var sign = bytes[0];
        string recvData = "";
        if (sign == (byte)WebProtocol.PackSign.SendAddr)
        {
            var resData = WebProtocol.SendAddrPackage.FromBytes(bytes);
            if (resData.IsSuccessful)
                recvData = resData.Value.ToString();
            else
                recvData = resData.Error.ToString();
        }
        else if (sign == (byte)WebProtocol.PackSign.SendData) { }
        else if (sign == (byte)WebProtocol.PackSign.RecvData) { }
        else if (sign == (byte)WebProtocol.PackSign.RecvResp) { }
        else
        {
            recvData = Encoding.ASCII.GetString(bytes, 0, bytes.Length);
        }

        Console.WriteLine($"Receive Data from {remoteStr} at {nowtime.ToString()}:");
        Console.WriteLine($"Original Data: {BitConverter.ToString(bytes).Replace("-", " ")}");
        if (recvData.Length != 0) Console.WriteLine(recvData);
        Console.WriteLine();

        listener.BeginReceive(new AsyncCallback(ReceiveHandler), null);
    }

    /// <summary>
    /// Start UDP Server
    /// </summary>
    /// <returns>None</returns>
    public void Start()
    {
        try
        {
            listener.BeginReceive(new AsyncCallback(ReceiveHandler), null);
        }
        catch (Exception e)
        {
            Console.WriteLine(e.ToString());
        }
        finally
        {

        }
    }

    /// <summary>
    /// Close UDP Server
    /// </summary>
    /// <returns>None</returns>
    public void Stop()
    {
        listener.Close();
    }
}

